# -*- coding: utf-8 -*-
"""Define project metadata
"""

__package_name__ = 'pydash'
__description__ = ('The kitchen sink of Python utility libraries '
                   'for doing "stuff" in a functional way. '
                   'Based on the Lo-Dash Javascript library.')
__url__ = 'https://github.com/dgilland/pydash'
__version__ = '4.5.0'
__author__ = 'Derrick Gilland'
__email__ = 'dgilland@gmail.com'
__license__ = 'MIT License'
